"""Asynchronous Python client for Overseerr."""

from collections.abc import AsyncGenerator, Generator

import aiohttp
from aioresponses import aioresponses
import pytest

from python_overseerr import OverseerrClient
from syrupy import SnapshotAssertion

from .syrupy import OverseerrSnapshotExtension


@pytest.fixture(name="snapshot")
def snapshot_assertion(snapshot: SnapshotAssertion) -> SnapshotAssertion:
    """Return snapshot assertion fixture with the Overseerr extension."""
    return snapshot.use_extension(OverseerrSnapshotExtension)


@pytest.fixture
async def client() -> AsyncGenerator[OverseerrClient, None]:
    """Return a Overseerr client."""
    async with (
        aiohttp.ClientSession() as session,
        OverseerrClient(
            "192.168.0.30",
            443,
            "key",
            session=session,
        ) as overseerr_client,
    ):
        yield overseerr_client


@pytest.fixture(name="responses")
def aioresponses_fixture() -> Generator[aioresponses, None, None]:
    """Return aioresponses fixture."""
    with aioresponses() as mocked_responses:
        yield mocked_responses



"""Constants for tests."""

from importlib import metadata

MOCK_HOST = "192.168.0.30"

MOCK_URL = f"https://{MOCK_HOST}/api/v1"
version = metadata.version("python_overseerr")

HEADERS = {
    "User-Agent": f"PythonOverseerr/{version}",
    "Accept": "application/json",
    "X-Api-Key": "key",
}



"""Asynchronous Python client for Overseerr."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from importlib import metadata
import socket
from typing import TYPE_CHECKING, Any, Literal
from urllib.parse import quote

from aiohttp import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_DELETE, METH_GET, METH_POST, METH_PUT
from yarl import URL

from .exceptions import OverseerrAuthenticationError, OverseerrConnectionError
from .models import (
    Issue,
    IssueCount,
    IssueFilterStatus,
    IssueResponse,
    IssueSortStatus,
    IssueStatus,
    IssueType,
    MediaType,
    MovieDetails,
    NotificationType,
    RequestCount,
    RequestFilterStatus,
    RequestResponse,
    RequestSortStatus,
    RequestWithMedia,
    Result,
    SearchResult,
    Status,
    TVDetails,
    WatchlistEntry,
    WatchlistResponse,
    WebhookNotificationConfig,
)

if TYPE_CHECKING:
    from typing_extensions import Self


VERSION = metadata.version(__package__)


@dataclass
class OverseerrClient:
    """Main class for handling connections with Overseerr."""

    host: str
    port: int
    api_key: str
    ssl: bool = True
    session: ClientSession | None = None
    request_timeout: int = 10
    _close_session: bool = False

    async def _request(
        self,
        method: str,
        uri: str,
        *,
        params: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
    ) -> str:
        """Handle a request to Overseerr."""
        url = URL.build(
            host=self.host, port=self.port, scheme="https" if self.ssl else "http"
        ).joinpath(f"api/v1/{uri}")

        headers = {
            "User-Agent": f"PythonOverseerr/{VERSION}",
            "Accept": "application/json",
            "X-Api-Key": self.api_key,
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    params=params,
                    headers=headers,
                    json=data,
                )
        except asyncio.TimeoutError as exception:
            msg = "Timeout occurred while connecting to the service"
            raise OverseerrConnectionError(msg) from exception
        except (
            ClientError,
            ClientResponseError,
            socket.gaierror,
        ) as exception:
            msg = "Error occurred while communicating with the service"
            raise OverseerrConnectionError(msg) from exception

        if response.status == 403:
            msg = "Invalid API key"
            raise OverseerrAuthenticationError(msg)

        if response.status >= 400:
            content_type = response.headers.get("Content-Type", "")
            text = await response.text()
            msg = "Unexpected response from Overseerr"
            raise OverseerrConnectionError(
                msg,
                {"Content-Type": content_type, "response": text},
            )

        return await response.text()

    async def get_request_count(self) -> RequestCount:
        """Get request count from Overseerr."""
        response = await self._request(METH_GET, "request/count")
        return RequestCount.from_json(response)

    async def get_issue_count(self) -> IssueCount:
        """Get issue count from Overseerr."""
        response = await self._request(METH_GET, "issue/count")
        return IssueCount.from_json(response)

    async def get_status(self) -> Status:
        """Get status from Overseerr."""
        response = await self._request(METH_GET, "status")
        return Status.from_json(response)

    async def search(self, keyword: str) -> list[Result]:
        """Search for media in Overseerr."""
        response = await self._request(
            METH_GET, "search", params={"query": quote(keyword)}
        )
        return SearchResult.from_json(response).results

    async def get_webhook_notification_config(self) -> WebhookNotificationConfig:
        """Get webhook notification config from Overseerr."""
        response = await self._request(METH_GET, "settings/notifications/webhook")
        return WebhookNotificationConfig.from_json(response)

    async def get_requests(
        self,
        status: RequestFilterStatus | None = None,
        sort: RequestSortStatus | None = None,
        requested_by: int | None = None,
    ) -> list[RequestWithMedia]:
        """Get requests from Overseerr."""
        params: dict[str, Any] = {}
        if status:
            params["filter"] = status
        if sort:
            params["sort"] = sort
        if requested_by:
            params["requestedBy"] = requested_by
        response = await self._request(METH_GET, "request", params=params)
        return RequestResponse.from_json(response).results

    async def create_request(
        self,
        media_type: MediaType,
        tmdb_id: int,
        seasons: list[int] | Literal["all"] | None = None,
    ) -> RequestWithMedia:
        """Create a request in Overseerr."""
        data = {"mediaType": media_type, "mediaId": tmdb_id}
        if seasons:
            data["seasons"] = seasons
        response = await self._request(METH_POST, "request", data=data)
        return RequestWithMedia.from_json(response)

    async def get_issues(
        self,
        status: IssueFilterStatus | None = None,
        sort: IssueSortStatus | None = None,
        requested_by: int | None = None,
    ) -> list[Issue]:
        """Get issues from Overseerr."""
        params: dict[str, Any] = {}
        if status:
            params["filter"] = status
        if sort:
            params["sort"] = sort
        if requested_by:
            params["requestedBy"] = requested_by
        response = await self._request(METH_GET, "issue", params=params)
        return IssueResponse.from_json(response).results

    async def get_issue(self, issue_id: int) -> Issue:
        """Get a single issue from Overseerr."""
        response = await self._request(METH_GET, f"issue/{issue_id}")
        return Issue.from_json(response)

    async def create_issue(
        self,
        issue_type: IssueType,
        message: str,
        media_id: int,
        problem_season: int = 0,
        problem_episode: int = 0,
    ) -> Issue:
        """Create a new issue in Overseerr."""
        data = {
            "issueType": issue_type.value,
            "message": message,
            "mediaId": media_id,
            "problemSeason": problem_season,
            "problemEpisode": problem_episode,
        }
        response = await self._request(METH_POST, "issue", data=data)
        return Issue.from_json(response)

    async def update_issue(
        self,
        issue_id: int,
        *,
        status: IssueStatus | None = None,
        message: str | None = None,
    ) -> Issue:
        """Update an existing issue in Overseerr."""
        data: dict[str, Any] = {}
        if status is not None:
            data["status"] = status.value
        if message is not None:
            data["message"] = message
        response = await self._request(METH_PUT, f"issue/{issue_id}", data=data)
        return Issue.from_json(response)

    async def delete_issue(self, issue_id: int) -> None:
        """Delete an issue from Overseerr."""
        await self._request(METH_DELETE, f"issue/{issue_id}")

    async def get_movie_details(self, identifier: int) -> MovieDetails:
        """Get movie details from Overseerr."""
        response = await self._request(METH_GET, f"movie/{identifier}")
        return MovieDetails.from_json(response)

    async def get_tv_details(self, identifier: int) -> TVDetails:
        """Get tv details from Overseerr."""
        response = await self._request(METH_GET, f"tv/{identifier}")
        return TVDetails.from_json(response)

    async def get_watchlist(self) -> list[WatchlistEntry]:
        """Get watchlist from Overseerr."""
        response = await self._request(METH_GET, "discover/watchlist")
        return WatchlistResponse.from_json(response).results

    async def test_webhook_notification_config(
        self, webhook_url: str, json_payload: str
    ) -> bool:
        """Test webhook notification config with Overseerr."""
        try:
            await self._request(
                METH_POST,
                "settings/notifications/webhook/test",
                data={
                    "enabled": True,
                    "types": NotificationType.REQUEST_PENDING_APPROVAL,
                    "options": {"webhookUrl": webhook_url, "jsonPayload": json_payload},
                },
            )
        except OverseerrConnectionError:
            return False
        return True

    async def set_webhook_notification_config(
        self,
        *,
        enabled: bool,
        types: NotificationType,
        webhook_url: str,
        json_payload: str,
    ) -> None:
        """Get webhook notification config from Overseerr."""
        await self._request(
            METH_POST,
            "settings/notifications/webhook",
            data={
                "enabled": enabled,
                "types": types,
                "options": {"webhookUrl": webhook_url, "jsonPayload": json_payload},
            },
        )

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The OverseerrClient object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()



"""Asynchronous Python client for Overseerr."""


class OverseerrError(Exception):
    """Generic exception."""


class OverseerrConnectionError(OverseerrError):
    """Overseerr connection exception."""


class OverseerrAuthenticationError(OverseerrError):
    """Overseerr authentication exception."""



"""Asynchronous Python client for Overseerr."""

from python_overseerr.exceptions import (
    OverseerrAuthenticationError,
    OverseerrConnectionError,
    OverseerrError,
)
from python_overseerr.models import (
    TV,
    Genre,
    Issue,
    IssueComment,
    IssueCount,
    IssueFilterStatus,
    IssueResponse,
    IssueSortStatus,
    IssueStatus,
    IssueType,
    Keyword,
    MediaInfo,
    MediaInfoWithRequests,
    MediaStatus,
    MediaType,
    Movie,
    MovieDetails,
    NotificationConfig,
    NotificationType,
    Person,
    Request,
    RequestCount,
    RequestFilterStatus,
    RequestResponse,
    RequestSortStatus,
    RequestStatus,
    Result,
    SearchResult,
    Status,
    TVDetails,
    User,
    WatchlistEntry,
    WatchlistResponse,
    WebhookNotificationConfig,
    WebhookNotificationOptions,
)
from python_overseerr.overseerr import OverseerrClient

__all__ = [
    "TV",
    "Genre",
    "Issue",
    "IssueComment",
    "IssueCount",
    "IssueFilterStatus",
    "IssueResponse",
    "IssueSortStatus",
    "IssueStatus",
    "IssueType",
    "Keyword",
    "MediaInfo",
    "MediaInfoWithRequests",
    "MediaStatus",
    "MediaType",
    "Movie",
    "MovieDetails",
    "NotificationConfig",
    "NotificationType",
    "OverseerrAuthenticationError",
    "OverseerrClient",
    "OverseerrConnectionError",
    "OverseerrError",
    "Person",
    "Request",
    "RequestCount",
    "RequestFilterStatus",
    "RequestResponse",
    "RequestSortStatus",
    "RequestStatus",
    "Result",
    "SearchResult",
    "Status",
    "TVDetails",
    "User",
    "WatchlistEntry",
    "WatchlistResponse",
    "WebhookNotificationConfig",
    "WebhookNotificationOptions",
]



"""Models for Overseerr."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date, datetime  # noqa: TC003
from enum import IntEnum, IntFlag, StrEnum
from typing import Annotated

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import Discriminator


@dataclass
class RequestCount(DataClassORJSONMixin):
    """Request count model."""

    total: int
    movie: int
    tv: int
    pending: int
    approved: int
    declined: int
    processing: int
    available: int


@dataclass
class IssueCount(DataClassORJSONMixin):
    """Issue count model."""

    total: int
    video: int
    audio: int
    subtitles: int
    others: int
    open: int
    closed: int


@dataclass
class Status(DataClassORJSONMixin):
    """Status model."""

    version: str
    update_available: bool = field(metadata=field_options(alias="updateAvailable"))
    commits_behind: int = field(metadata=field_options(alias="commitsBehind"))
    restart_required: bool = field(metadata=field_options(alias="restartRequired"))


class MediaStatus(IntEnum):
    """Media status enum."""

    UNKNOWN = 1
    PENDING = 2
    PROCESSING = 3
    PARTIALLY_AVAILABLE = 4
    AVAILABLE = 5


@dataclass
class MediaInfo(DataClassORJSONMixin):
    """Media info model."""

    id: int
    tmdb_id: int | None = field(metadata=field_options(alias="tmdbId"))
    tvdb_id: int | None = field(metadata=field_options(alias="tvdbId"))
    imdb_id: str | None = field(metadata=field_options(alias="imdbId"))
    media_type: MediaType = field(metadata=field_options(alias="mediaType"))
    status: MediaStatus
    created_at: datetime = field(metadata=field_options(alias="createdAt"))
    updated_at: datetime = field(metadata=field_options(alias="updatedAt"))


class ResultMediaType(StrEnum):
    """Media type enum."""

    MOVIE = "movie"
    TV = "tv"
    PERSON = "person"


class MediaType(StrEnum):
    """Media type enum."""

    MOVIE = "movie"
    TV = "tv"


@dataclass
class Result(DataClassORJSONMixin):
    """Result model."""

    id: int
    mediaType: ResultMediaType  # noqa: N815 # pylint: disable=invalid-name
    media_type: ResultMediaType = field(metadata=field_options(alias="mediaType"))


@dataclass
class Movie(Result):
    """Movie result model."""

    mediaType = ResultMediaType.MOVIE  # noqa: N815 # pylint: disable=invalid-name
    original_language: str = field(metadata=field_options(alias="originalLanguage"))
    original_title: str = field(metadata=field_options(alias="originalTitle"))
    overview: str
    popularity: float
    title: str
    adult: bool
    media_info: MediaInfo | None = field(
        metadata=field_options(alias="mediaInfo"), default=None
    )


@dataclass
class TV(Result):
    """TV result model."""

    mediaType = ResultMediaType.TV  # noqa: N815 # pylint: disable=invalid-name
    first_air_date: date = field(metadata=field_options(alias="firstAirDate"))
    name: str
    original_language: str = field(metadata=field_options(alias="originalLanguage"))
    original_name: str = field(metadata=field_options(alias="originalName"))
    overview: str
    popularity: float
    media_info: MediaInfo | None = field(
        metadata=field_options(alias="mediaInfo"), default=None
    )


@dataclass
class Person(Result):
    """Person result model."""

    mediaType = ResultMediaType.PERSON  # noqa: N815 # pylint: disable=invalid-name
    name: str
    popularity: float
    known_for: list[Result] = field(metadata=field_options(alias="knownFor"))
    adult: bool


@dataclass
class SearchResult(DataClassORJSONMixin):
    """Search result model."""

    results: list[
        Annotated[Result, Discriminator(field="mediaType", include_subtypes=True)]
    ]


class NotificationType(IntFlag):
    """Webhook notification type enum."""

    REQUEST_PENDING_APPROVAL = 2
    REQUEST_APPROVED = 4
    REQUEST_AVAILABLE = 8
    REQUEST_PROCESSING_FAILED = 16
    REQUEST_DECLINED = 64
    REQUEST_AUTOMATICALLY_APPROVED = 128
    ISSUE_REPORTED = 256
    ISSUE_COMMENTED = 512
    ISSUE_RESOLVED = 1024
    ISSUE_REOPENED = 2048


@dataclass
class NotificationConfig(DataClassORJSONMixin):
    """Webhook config model."""

    enabled: bool
    types: NotificationType


@dataclass
class WebhookNotificationOptions:
    """Webhook notification options model."""

    json_payload: str = field(metadata=field_options(alias="jsonPayload"))
    webhook_url: str = field(metadata=field_options(alias="webhookUrl"))


@dataclass
class WebhookNotificationConfig(NotificationConfig):
    """Webhook config model."""

    options: WebhookNotificationOptions


@dataclass
class User(DataClassORJSONMixin):
    """User model."""

    id: int
    email: str
    avatar: str
    username: str | None
    plex_username: str | None = field(metadata=field_options(alias="plexUsername"))
    plex_id: int | None = field(metadata=field_options(alias="plexId"))
    movie_quota_limit: int | None = field(
        metadata=field_options(alias="movieQuotaLimit")
    )
    movie_quota_days: int | None = field(metadata=field_options(alias="movieQuotaDays"))
    tv_quota_limit: int | None = field(metadata=field_options(alias="tvQuotaLimit"))
    tv_quota_days: int | None = field(metadata=field_options(alias="tvQuotaDays"))
    created_at: datetime = field(metadata=field_options(alias="createdAt"))
    updated_at: datetime = field(metadata=field_options(alias="updatedAt"))
    request_count: int = field(metadata=field_options(alias="requestCount"))
    display_name: str = field(metadata=field_options(alias="displayName"))
    permissions: int
    recovery_link_expiration_date: date | None = field(
        metadata=field_options(alias="recoveryLinkExpirationDate")
    )
    user_type: int | None = field(metadata=field_options(alias="userType"))
    settings: UserSettings | None = None
    warnings: list[str] = field(default_factory=list)
    jellyfin_username: str | None = field(
        metadata=field_options(alias="jellyfinUsername"), default=None
    )
    jellyfin_user_id: str | None = field(
        metadata=field_options(alias="jellyfinUserId"), default=None
    )
    avatar_e_tag: str | None = field(
        metadata=field_options(alias="avatarETag"), default=None
    )
    avatar_version: str | None = field(
        metadata=field_options(alias="avatarVersion"), default=None
    )


@dataclass
class UserSettings(DataClassORJSONMixin):
    """User settings model."""

    discord_id: str | None = field(metadata=field_options(alias="discordId"))
    id: str
    locale: str
    notification_types: NotificationTypes = field(
        metadata=field_options(alias="notificationTypes")
    )
    original_language: str | None = field(
        metadata=field_options(alias="originalLanguage")
    )
    pgp_key: str | None = field(metadata=field_options(alias="pgpKey"))
    pushbullet_access_token: str | None = field(
        metadata=field_options(alias="pushbulletAccessToken")
    )
    pushover_application_token: str | None = field(
        metadata=field_options(alias="pushoverApplicationToken")
    )
    pushover_sound: str | None = field(metadata=field_options(alias="pushoverSound"))
    pushover_user_key: str | None = field(
        metadata=field_options(alias="pushoverUserKey")
    )
    telegram_chat_id: str | None = field(metadata=field_options(alias="telegramChatId"))
    telegram_send_silently: str | None = field(
        metadata=field_options(alias="telegramSendSilently")
    )
    watchlist_sync_movies: str | None = field(
        metadata=field_options(alias="watchlistSyncMovies")
    )
    watchlist_sync_tv: str | None = field(
        metadata=field_options(alias="watchlistSyncTv")
    )
    discover_region: str | None = field(
        metadata=field_options(alias="discoverRegion"), default=None
    )
    streaming_region: str | None = field(
        metadata=field_options(alias="streamingRegion"), default=None
    )
    telegram_message_thread_id: str | None = field(
        metadata=field_options(alias="telegramMessageThreadId"), default=None
    )


@dataclass
class NotificationTypes(DataClassORJSONMixin):
    """Notification types model."""

    email: int | None = field(default=None)
    discord: int | None = field(default=None)
    pushbullet: int | None = field(default=None)
    pushover: int | None = field(default=None)
    slack: int | None = field(default=None)
    telegram: int | None = field(default=None)
    webhook: int | None = field(default=None)
    webpush: int | None = field(default=None)


class RequestFilterStatus(StrEnum):
    """Request filter status enum."""

    ALL = "all"
    APPROVED = "approved"
    AVAILABLE = "available"
    PENDING = "pending"
    PROCESSING = "processing"
    UNAVAILABLE = "unavailable"
    FAILED = "failed"


class RequestSortStatus(StrEnum):
    """Request sort status enum."""

    ADDED = "added"
    MODIFIED = "modified"


class RequestStatus(IntEnum):
    """Request status enum."""

    PENDING_APPROVAL = 1
    APPROVED = 2
    DECLINED = 3
    COMPLETED = 5


@dataclass
class Request(DataClassORJSONMixin):
    """Request model."""

    id: int
    status: RequestStatus
    created_at: datetime = field(metadata=field_options(alias="createdAt"))
    updated_at: datetime = field(metadata=field_options(alias="updatedAt"))
    is4k: bool
    requested_by: User = field(metadata=field_options(alias="requestedBy"))
    season_count: int | None = field(
        metadata=field_options(alias="seasonCount"), default=None
    )
    modified_by: User | None = field(
        metadata=field_options(alias="modifiedBy"), default=None
    )


@dataclass(kw_only=True)
class RequestWithMedia(Request):
    """Request with media model."""

    media: MediaInfo


@dataclass
class MediaInfoWithRequests(MediaInfo):
    """Media info with requests model."""

    requests: list[Request]


@dataclass
class RequestResponse(DataClassORJSONMixin):
    """Request response model."""

    results: list[RequestWithMedia]


class IssueFilterStatus(StrEnum):
    """Issue filter status enum."""

    ALL = "all"
    OPEN = "open"
    RESOLVED = "resolved"


class IssueSortStatus(StrEnum):
    """Issue sort status enum."""

    ADDED = "added"
    MODIFIED = "modified"


class IssueStatus(IntEnum):
    """Issue status enum."""

    OPEN = 1
    RESOLVED = 2


class IssueType(IntEnum):
    """Issue status enum."""

    VIDEO = 1
    AUDIO = 2
    SUBTITLE = 3
    OTHER = 4


@dataclass
class IssueComment(DataClassORJSONMixin):
    """Issue comment model."""

    id: int
    message: str
    created_at: datetime = field(metadata=field_options(alias="createdAt"))
    user: User | None = None


@dataclass
class Issue(DataClassORJSONMixin):
    """Issue model."""

    id: int
    issue_type: IssueType = field(metadata=field_options(alias="issueType"))
    status: IssueStatus
    problem_season: int = field(metadata=field_options(alias="problemSeason"))
    problem_episode: int = field(metadata=field_options(alias="problemEpisode"))
    created_at: datetime = field(metadata=field_options(alias="createdAt"))
    updated_at: datetime = field(metadata=field_options(alias="updatedAt"))
    created_by: User = field(metadata=field_options(alias="createdBy"))
    media: MediaInfo
    modified_by: User | None = field(
        metadata=field_options(alias="modifiedBy"), default=None
    )
    comments: list[IssueComment] = field(default_factory=list)


@dataclass
class IssueResponse(DataClassORJSONMixin):
    """Issue response model."""

    results: list[Issue]


@dataclass
class Genre:
    """Genre model."""

    id: int
    name: str


@dataclass
class Keyword:
    """Keyword model."""

    id: int
    name: str


@dataclass
class MovieDetails(DataClassORJSONMixin):
    """Movie details model."""

    id: int
    adult: bool
    budget: int
    genres: list[Genre]
    original_language: str = field(metadata=field_options(alias="originalLanguage"))
    original_title: str = field(metadata=field_options(alias="originalTitle"))
    popularity: float
    release_date: date = field(metadata=field_options(alias="releaseDate"))
    revenue: int
    title: str
    vote_average: float = field(metadata=field_options(alias="voteAverage"))
    vote_count: int = field(metadata=field_options(alias="voteCount"))
    imdb_id: str | None = field(metadata=field_options(alias="imdbId"))
    overview: str
    runtime: int
    tagline: str
    keywords: list[Keyword]
    media_info: MediaInfoWithRequests | None = field(
        metadata=field_options(alias="mediaInfo"), default=None
    )


@dataclass
class Season(DataClassORJSONMixin):
    """Season model."""

    id: int
    name: str
    overview: str
    season_number: int = field(metadata=field_options(alias="seasonNumber"))
    episode_count: int = field(metadata=field_options(alias="episodeCount"))
    air_date: date | None = field(metadata=field_options(alias="airDate"))
    poster_path: str = field(metadata=field_options(alias="posterPath"))


@dataclass
class Episode(DataClassORJSONMixin):
    """Episode model."""

    id: int
    name: str
    overview: str
    episode_number: int = field(metadata=field_options(alias="episodeNumber"))
    still_path: str = field(metadata=field_options(alias="stillPath"))
    air_date: date | None = field(metadata=field_options(alias="airDate"), default=None)


@dataclass
class TVDetails(DataClassORJSONMixin):
    """TV details model."""

    id: int
    first_air_date: date = field(metadata=field_options(alias="firstAirDate"))
    genres: list[Genre]
    languages: list[str]
    last_air_date: date = field(metadata=field_options(alias="lastAirDate"))
    name: str
    number_of_episodes: int = field(metadata=field_options(alias="numberOfEpisodes"))
    number_of_seasons: int = field(metadata=field_options(alias="numberOfSeasons"))
    original_language: str = field(metadata=field_options(alias="originalLanguage"))
    original_name: str = field(metadata=field_options(alias="originalName"))
    tagline: str
    overview: str
    popularity: float
    seasons: list[Season]
    last_episode_to_air: Episode = field(
        metadata=field_options(alias="lastEpisodeToAir")
    )
    keywords: list[Keyword]
    media_info: MediaInfoWithRequests | None = field(
        metadata=field_options(alias="mediaInfo"), default=None
    )
    next_episode_to_air: Episode | None = field(
        metadata=field_options(alias="nextEpisodeToAir"), default=None
    )


@dataclass
class WatchlistEntry(DataClassORJSONMixin):
    """Watchlist entry model."""

    rating_key: str = field(metadata=field_options(alias="ratingKey"))
    title: str
    media_type: MediaType = field(metadata=field_options(alias="mediaType"))
    tmdb_id: int | None = field(metadata=field_options(alias="tmdbId"))


@dataclass
class WatchlistResponse(DataClassORJSONMixin):
    """Watchlist response model."""

    results: list[WatchlistEntry]



"""Tests for the client."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any

import aiohttp
from aiohttp import ClientError
from aiohttp.hdrs import METH_DELETE, METH_GET, METH_POST, METH_PUT
from aioresponses import CallbackResult, aioresponses
import pytest

from python_overseerr import MediaType, OverseerrClient
from python_overseerr.exceptions import (
    OverseerrAuthenticationError,
    OverseerrConnectionError,
    OverseerrError,
)
from python_overseerr.models import (
    IssueStatus,
    IssueType,
    NotificationType,
    RequestFilterStatus,
    RequestSortStatus,
)
from tests import load_fixture
from tests.const import HEADERS, MOCK_URL

if TYPE_CHECKING:
    from syrupy import SnapshotAssertion

SERVICES = ["overseerr", "seerr"]


async def test_putting_in_own_session(
    responses: aioresponses,
) -> None:
    """Test putting in own session."""
    responses.get(
        f"{MOCK_URL}/request/count",
        status=200,
        body=load_fixture("overseerr/request_count.json"),
    )
    async with aiohttp.ClientSession() as session:
        overseerr = OverseerrClient("192.168.0.30", 443, "abc", session=session)
        await overseerr.get_request_count()
        assert overseerr.session is not None
        assert not overseerr.session.closed
        await overseerr.close()
        assert not overseerr.session.closed


async def test_creating_own_session(
    responses: aioresponses,
) -> None:
    """Test creating own session."""
    responses.get(
        f"{MOCK_URL}/request/count",
        status=200,
        body=load_fixture("overseerr/request_count.json"),
    )
    overseerr = OverseerrClient("192.168.0.30", 443, "abc")
    await overseerr.get_request_count()
    assert overseerr.session is not None
    assert not overseerr.session.closed
    await overseerr.close()
    assert overseerr.session.closed


async def test_unexpected_server_response(
    responses: aioresponses,
    client: OverseerrClient,
) -> None:
    """Test handling unexpected response."""
    responses.get(
        f"{MOCK_URL}/request/count",
        status=404,
        headers={"Content-Type": "plain/text"},
        body="Yes",
    )
    with pytest.raises(OverseerrError):
        await client.get_request_count()


async def test_timeout(
    responses: aioresponses,
) -> None:
    """Test request timeout."""

    # Faking a timeout by sleeping
    async def response_handler(_: str, **_kwargs: Any) -> CallbackResult:
        """Response handler for this test."""
        await asyncio.sleep(2)
        return CallbackResult(body="Goodmorning!")

    responses.get(
        f"{MOCK_URL}/request/count",
        callback=response_handler,
    )
    async with OverseerrClient(
        "192.168.0.30",
        443,
        "abc",
        request_timeout=1,
    ) as overseerr:
        with pytest.raises(OverseerrConnectionError):
            await overseerr.get_request_count()


async def test_client_error(
    client: OverseerrClient,
    responses: aioresponses,
) -> None:
    """Test client error."""

    async def response_handler(_: str, **_kwargs: Any) -> CallbackResult:
        """Response handler for this test."""
        raise ClientError

    responses.get(
        f"{MOCK_URL}/request/count",
        callback=response_handler,
    )
    with pytest.raises(OverseerrConnectionError):
        await client.get_request_count()


@pytest.mark.parametrize("service", SERVICES)
async def test_authentication_error(
    client: OverseerrClient,
    responses: aioresponses,
    service: str,
) -> None:
    """Test authentication error."""
    responses.get(
        f"{MOCK_URL}/request/count",
        status=403,
        body=load_fixture(f"{service}/no_access.json"),
    )
    with pytest.raises(OverseerrAuthenticationError):
        await client.get_request_count()


@pytest.mark.parametrize("service", SERVICES)
@pytest.mark.parametrize(
    ("endpoint", "fixture", "method"),
    [
        ("request/count", "request_count.json", "get_request_count"),
        ("status", "status.json", "get_status"),
        (
            "settings/notifications/webhook",
            "webhook_config.json",
            "get_webhook_notification_config",
        ),
        (
            "discover/watchlist",
            "watchlist.json",
            "get_watchlist",
        ),
        ("issue/count", "issue_count.json", "get_issue_count"),
    ],
    ids=[
        "request_count",
        "status",
        "webhook_config",
        "watchlist",
        "issue_count",
    ],
)
async def test_data_retrieval(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    endpoint: str,
    fixture: str,
    method: str,
    service: str,
) -> None:
    """Test data retrieval."""
    responses.get(
        f"{MOCK_URL}/{endpoint}",
        status=200,
        body=load_fixture(f"{service}/{fixture}"),
    )
    assert await getattr(client, method)() == snapshot
    responses.assert_called_once_with(
        f"{MOCK_URL}/{endpoint}",
        METH_GET,
        headers=HEADERS,
        params=None,
        json=None,
    )


@pytest.mark.parametrize("service", SERVICES)
@pytest.mark.parametrize(
    "fixture",
    [
        "search_1.json",
        "search_2.json",
    ],
)
async def test_search(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    fixture: str,
    service: str,
) -> None:
    """Test searching for media."""
    responses.get(
        f"{MOCK_URL}/search?query=frosty",
        status=200,
        body=load_fixture(f"{service}/{fixture}"),
    )
    assert await client.search("frosty") == snapshot
    responses.assert_called_once_with(
        f"{MOCK_URL}/search",
        METH_GET,
        headers=HEADERS,
        params={"query": "frosty"},
        json=None,
    )


async def test_search_with_encoded_spaces(
    responses: aioresponses, client: OverseerrClient
) -> None:
    """Test searching for media."""
    responses.get(
        f"{MOCK_URL}/search?query=frosty%20the%20snowman",
        status=200,
        body=load_fixture("overseerr/search_1.json"),
    )
    assert await client.search("frosty the snowman")
    responses.assert_called_once_with(
        f"{MOCK_URL}/search",
        METH_GET,
        headers=HEADERS,
        params={"query": "frosty%20the%20snowman"},
        json=None,
    )


async def test_setting_webhook_configuration(
    responses: aioresponses,
    client: OverseerrClient,
) -> None:
    """Test setting webhook configuration."""
    responses.post(
        f"{MOCK_URL}/settings/notifications/webhook",
        status=200,
    )
    await client.set_webhook_notification_config(
        enabled=True,
        types=NotificationType.REQUEST_APPROVED,
        webhook_url="http://localhost",
        json_payload="{}",
    )
    responses.assert_called_once_with(
        f"{MOCK_URL}/settings/notifications/webhook",
        METH_POST,
        headers=HEADERS,
        params=None,
        json={
            "enabled": True,
            "types": 4,
            "options": {
                "webhookUrl": "http://localhost",
                "jsonPayload": "{}",
            },
        },
    )


async def test_webhook_config_test(
    responses: aioresponses,
    client: OverseerrClient,
) -> None:
    """Test setting webhook configuration."""
    responses.post(
        f"{MOCK_URL}/settings/notifications/webhook/test",
        status=204,
    )
    assert (
        await client.test_webhook_notification_config(
            webhook_url="http://localhost",
            json_payload="{}",
        )
        is True
    )
    responses.assert_called_once_with(
        f"{MOCK_URL}/settings/notifications/webhook/test",
        METH_POST,
        headers=HEADERS,
        params=None,
        json={
            "enabled": True,
            "types": 2,
            "options": {
                "webhookUrl": "http://localhost",
                "jsonPayload": "{}",
            },
        },
    )


async def test_failing_webhook_config_test(
    responses: aioresponses,
    client: OverseerrClient,
) -> None:
    """Test setting webhook configuration."""
    responses.post(
        f"{MOCK_URL}/settings/notifications/webhook/test",
        status=500,
        body='{"message": "Failed to send webhook notification."}',
    )
    assert (
        await client.test_webhook_notification_config(
            webhook_url="http://localhost",
            json_payload="{}",
        )
        is False
    )
    responses.assert_called_once_with(
        f"{MOCK_URL}/settings/notifications/webhook/test",
        METH_POST,
        headers=HEADERS,
        params=None,
        json={
            "enabled": True,
            "types": 2,
            "options": {
                "webhookUrl": "http://localhost",
                "jsonPayload": "{}",
            },
        },
    )


@pytest.mark.parametrize("service", SERVICES)
async def test_fetching_requests(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    service: str,
) -> None:
    """Test fetching requests."""
    responses.get(
        f"{MOCK_URL}/request",
        status=200,
        body=load_fixture(f"{service}/request.json"),
    )
    assert await client.get_requests() == snapshot
    responses.assert_called_once_with(
        f"{MOCK_URL}/request", METH_GET, headers=HEADERS, params={}, json=None
    )


@pytest.mark.parametrize(
    ("kwargs", "params", "query_string"),
    [
        ({"status": RequestFilterStatus.ALL}, {"filter": "all"}, "filter=all"),
        ({"sort": RequestSortStatus.ADDED}, {"sort": "added"}, "sort=added"),
        ({"requested_by": 1}, {"requestedBy": 1}, "requestedBy=1"),
    ],
)
async def test_fetching_request_parameters(
    responses: aioresponses,
    client: OverseerrClient,
    kwargs: dict[str, Any],
    params: dict[str, Any],
    query_string: str,
) -> None:
    """Test fetching requests with parameters."""
    responses.get(
        f"{MOCK_URL}/request?{query_string}",
        status=200,
        body=load_fixture("overseerr/request.json"),
    )
    await client.get_requests(**kwargs)
    responses.assert_called_once_with(
        f"{MOCK_URL}/request", METH_GET, headers=HEADERS, params=params, json=None
    )


@pytest.mark.parametrize("service", SERVICES)
async def test_fetching_issues(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    service: str,
) -> None:
    """Test fetching issues."""
    responses.get(
        f"{MOCK_URL}/issue",
        status=200,
        body=load_fixture(f"{service}/issue.json"),
    )
    assert await client.get_issues() == snapshot
    responses.assert_called_once_with(
        f"{MOCK_URL}/issue", METH_GET, headers=HEADERS, params={}, json=None
    )


@pytest.mark.parametrize(
    ("kwargs", "params", "query_string"),
    [
        ({"status": RequestFilterStatus.ALL}, {"filter": "all"}, "filter=all"),
        ({"sort": RequestSortStatus.ADDED}, {"sort": "added"}, "sort=added"),
        ({"requested_by": 1}, {"requestedBy": 1}, "requestedBy=1"),
    ],
)
async def test_fetching_issue_parameters(
    responses: aioresponses,
    client: OverseerrClient,
    kwargs: dict[str, Any],
    params: dict[str, Any],
    query_string: str,
) -> None:
    """Test fetching issues with parameters."""
    responses.get(
        f"{MOCK_URL}/issue?{query_string}",
        status=200,
        body=load_fixture("overseerr/issue.json"),
    )
    await client.get_issues(**kwargs)
    responses.assert_called_once_with(
        f"{MOCK_URL}/issue", METH_GET, headers=HEADERS, params=params, json=None
    )


@pytest.mark.parametrize("service", SERVICES)
async def test_fetching_movie_details(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    service: str,
) -> None:
    """Test fetching movie details."""
    responses.get(
        f"{MOCK_URL}/movie/1156593",
        status=200,
        body=load_fixture(f"{service}/movie.json"),
    )
    assert await client.get_movie_details(1156593) == snapshot
    responses.assert_called_once_with(
        f"{MOCK_URL}/movie/1156593", METH_GET, headers=HEADERS, params=None, json=None
    )


@pytest.mark.parametrize("service", SERVICES)
async def test_fetching_tv_details(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    service: str,
) -> None:
    """Test fetching tv details."""
    responses.get(
        f"{MOCK_URL}/tv/249522",
        status=200,
        body=load_fixture(f"{service}/tv.json"),
    )
    assert await client.get_tv_details(249522) == snapshot
    responses.assert_called_once_with(
        f"{MOCK_URL}/tv/249522", METH_GET, headers=HEADERS, params=None, json=None
    )


@pytest.mark.parametrize("service", SERVICES)
@pytest.mark.parametrize(
    ("args", "fixture", "json"),
    [
        (
            (MediaType.MOVIE, 1156593),
            "create_movie_request.json",
            {"mediaType": "movie", "mediaId": 1156593},
        ),
        (
            (MediaType.TV, 249522, "all"),
            "create_tv_request.json",
            {"mediaType": "tv", "mediaId": 249522, "seasons": "all"},
        ),
        (
            (MediaType.TV, 249522, [1]),
            "create_tv_request.json",
            {"mediaType": "tv", "mediaId": 249522, "seasons": [1]},
        ),
    ],
)
async def test_creating_request(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    args: tuple[Any, ...],
    fixture: str,
    json: dict[str, Any],
    service: str,
) -> None:
    """Test creating a request."""
    responses.post(
        f"{MOCK_URL}/request",
        status=201,
        body=load_fixture(f"{service}/{fixture}"),
    )
    assert await client.create_request(*args) == snapshot
    responses.assert_called_once_with(
        f"{MOCK_URL}/request", METH_POST, headers=HEADERS, params=None, json=json
    )


@pytest.mark.parametrize("service", SERVICES)
async def test_fetching_single_issue(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    service: str,
) -> None:
    """Test fetching a single issue."""
    responses.get(
        f"{MOCK_URL}/issue/11",
        status=200,
        body=load_fixture(f"{service}/issue_single.json"),
    )
    assert await client.get_issue(11) == snapshot
    responses.assert_called_once_with(
        f"{MOCK_URL}/issue/11", METH_GET, headers=HEADERS, params=None, json=None
    )


@pytest.mark.parametrize("service", SERVICES)
async def test_creating_issue(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    service: str,
) -> None:
    """Test creating an issue."""
    responses.post(
        f"{MOCK_URL}/issue",
        status=201,
        body=load_fixture(f"{service}/issue_created.json"),
    )
    assert (
        await client.create_issue(
            issue_type=IssueType.VIDEO,
            message="Video playback not working",
            media_id=1156593,
            problem_season=0,
            problem_episode=0,
        )
        == snapshot
    )
    responses.assert_called_once_with(
        f"{MOCK_URL}/issue",
        METH_POST,
        headers=HEADERS,
        params=None,
        json={
            "issueType": 1,
            "message": "Video playback not working",
            "mediaId": 1156593,
            "problemSeason": 0,
            "problemEpisode": 0,
        },
    )


@pytest.mark.parametrize("service", SERVICES)
async def test_updating_issue_status(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    service: str,
) -> None:
    """Test updating issue status."""
    responses.put(
        f"{MOCK_URL}/issue/11",
        status=200,
        body=load_fixture(f"{service}/issue_updated.json"),
    )
    assert (
        await client.update_issue(
            issue_id=11,
            status=IssueStatus.RESOLVED,
        )
        == snapshot
    )
    responses.assert_called_once_with(
        f"{MOCK_URL}/issue/11",
        METH_PUT,
        headers=HEADERS,
        params=None,
        json={"status": 2},
    )


@pytest.mark.parametrize("service", SERVICES)
async def test_updating_issue_with_comment(
    responses: aioresponses,
    client: OverseerrClient,
    snapshot: SnapshotAssertion,
    service: str,
) -> None:
    """Test updating issue with comment."""
    responses.put(
        f"{MOCK_URL}/issue/11",
        status=200,
        body=load_fixture(f"{service}/issue_updated.json"),
    )
    assert (
        await client.update_issue(
            issue_id=11,
            status=IssueStatus.RESOLVED,
            message="Issue has been resolved",
        )
        == snapshot
    )
    responses.assert_called_once_with(
        f"{MOCK_URL}/issue/11",
        METH_PUT,
        headers=HEADERS,
        params=None,
        json={"status": 2, "message": "Issue has been resolved"},
    )


async def test_deleting_issue(
    responses: aioresponses,
    client: OverseerrClient,
) -> None:
    """Test deleting an issue."""
    responses.delete(
        f"{MOCK_URL}/issue/11",
        status=204,
    )
    await client.delete_issue(11)
    responses.assert_called_once_with(
        f"{MOCK_URL}/issue/11", METH_DELETE, headers=HEADERS, params=None, json=None
    )
